/***************************************************************************
 *                                                                         *
 *   copyright        : (C) 2011 Commissariat a l'Energie Atomique (CEA)   *
 *   author           : Gilles Mouchard                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "uart_lite.h"
#include "intc.h"
#include "io.h"
#include "mmu.h"
#include <inttypes.h>

#define UART_LITE_BASEADDR 0x84000000UL
#define UART_LITE_BASE_PHYSICAL_ADDR 0x84000000UL
#define UART_LITE_HIGH_PHYSICAL_ADDR 0x8400ffffUL

#define UART_LITE_RX_FIFO 0x0
#define UART_LITE_TX_FIFO 0x4
#define UART_LITE_STAT_REG 0x8
#define UART_LITE_CTRL_REG 0xc

#define UART_LITE_TX_FIFO_FULL 0x8
#define UART_LITE_RX_FIFO_VALID_DATA 0x1

#define UART_LITE_ENABLE_INTR 0x10

#define UART_LITE_IRQ 2

typedef struct
{
	uint32_t rx_fifo;
	uint32_t tx_fifo;
	uint32_t stat_reg;
	uint32_t ctrl_reg;
} uart_lite_t;

static uart_lite_t *uart_lite = 0;
static volatile int uart_lite_has_pending_event = 0;

void uart_lite_irq_handler()
{
	uart_lite_has_pending_event = 1; // The UART has incoming data ready or is ready to accept new data to send
}

void uart_lite_init()
{
	uint32_t byte_length = UART_LITE_HIGH_PHYSICAL_ADDR - UART_LITE_BASE_PHYSICAL_ADDR + 1;
	page_t page;
	
	page.valid = 1;
	page.ts = 0;
	page.base_effective_addr = UART_LITE_BASEADDR;
	page.tid = 0;
	page.base_physical_addr = UART_LITE_BASE_PHYSICAL_ADDR;
	page.size = get_page_size(byte_length);
	page.attr = ATTR_I | ATTR_SX | ATTR_SW | ATTR_SR;

	load_tlb(&page);

	uart_lite = (uart_lite_t *) UART_LITE_BASEADDR;
	
	register_irq_handler(UART_LITE_IRQ, uart_lite_irq_handler);
	enable_irq(UART_LITE_IRQ);
	write_big32(&uart_lite->ctrl_reg, read_big32(&uart_lite->ctrl_reg) | UART_LITE_ENABLE_INTR); // enable UART Lite interrupt
}

void PrintChar(char c)
{
	if(!uart_lite) return;
	while(read_big32(&uart_lite->stat_reg) & UART_LITE_TX_FIFO_FULL);
	
	write_big32(&uart_lite->tx_fifo, c);
}

void PrintString(char *s)
{
	static char prev_c = 0;
	if(!uart_lite) return;
	char c = *s;
	if(c)
	{
		do
		{
			if((c == '\n') && (prev_c != '\r')) PrintChar('\r'); // Add a carriage return before line feed
			else if((prev_c == '\r') && (c != '\n')) PrintChar('\n'); // Add a line feed after carriage return
			PrintChar(c);
			prev_c = c;
		}
		while((c = *++s));
	}
}

void PrintHex4(uint8_t v)
{
	PrintChar((v < 10) ? '0' + v : 'a' + (v - 10));
}

void PrintHex8(uint8_t v)
{
	PrintChar('0');
	PrintChar('x');
	PrintHex4((v >> 4) & 0xf);
	PrintHex4(v & 0xf);
}

void PrintHex16(uint16_t v)
{
	PrintChar('0');
	PrintChar('x');
	PrintHex4((v >> 12) & 0xf);
	PrintHex4((v >> 8) & 0xf);
	PrintHex4((v >> 4) & 0xf);
	PrintHex4(v & 0xf);
}

void PrintHex32(uint32_t v)
{
	PrintChar('0');
	PrintChar('x');
	PrintHex4((v >> 28) & 0xf);
	PrintHex4((v >> 24) & 0xf);
	PrintHex4((v >> 20) & 0xf);
	PrintHex4((v >> 16) & 0xf);
	PrintHex4((v >> 12) & 0xf);
	PrintHex4((v >> 8) & 0xf);
	PrintHex4((v >> 4) & 0xf);
	PrintHex4(v & 0xf);
}

void PrintUInt32(uint32_t v)
{
	uint32_t factor = 1000000000;
	int f = 0;
	do
	{
		uint32_t digit = v / factor;
		v -= digit * factor;
		
		if(digit) f = 1;
		if(f) PrintChar('0' + digit);
	}
	while(factor = factor / 10);
	
	if(!f) PrintChar('0');
}

void ReadString(char *s, unsigned int max_size)
{
	if(!uart_lite)
	{
		*s = 0;
		return;
	}
	while(!uart_lite_has_pending_event);
	
	write_big32(&uart_lite->ctrl_reg, read_big32(&uart_lite->ctrl_reg) & ~UART_LITE_ENABLE_INTR); // disable UART Lite interrupt to avoid concurent access to has_uart_lite_pending_data

	unsigned int max_length = max_size - 1;
	unsigned int i;
	char c = 0;
		
	for(i = 0; (read_big32(&uart_lite->stat_reg) & UART_LITE_RX_FIFO_VALID_DATA) && (i < max_length) && (c != '\n'); i++)
	{
		c = (char) read_big32(&uart_lite->rx_fifo); // acknowlegde uart lite when there's no more byte to read
		s[i] = c;
	}

	uart_lite_has_pending_event = (read_big32(&uart_lite->stat_reg) & UART_LITE_RX_FIFO_VALID_DATA) != 0;
	write_big32(&uart_lite->ctrl_reg, read_big32(&uart_lite->ctrl_reg) | UART_LITE_ENABLE_INTR); // re-enable UART Lite interrupt

	s[i] = 0; // finish string with a null character
}
